analyze_individual_network <- function(counts, threshold = NULL, n_pls_comp = 1,
                                       max_n_edges = 10000, make_plots = TRUE,
                                       verbose = TRUE) {
  network_name <- attr(counts, "source")
  
  if(verbose) cat("Analyzing network", network_name, "\n")
  # Determine which filters were used on the counts; used for labeling.
  filters <- ""
  n_filters <- length(attr(counts, "filters"))
  if(n_filters > 0) {
    filter_vector <- sapply(attr(counts, "filters"), c)
    for(i in 1:n_filters) {
      filters <- paste(filters, 
                       "_", names(filter_vector)[i], filter_vector[i], sep = "")
    }
    if(verbose) cat("\tFilters found:", filters, "\n")
  } else {
    if(verbose) cat("\tNo filters found\n")
  }
  
  network_title <- paste(network_name, filters, sep = "")
  
  if(verbose) cat("\tComputing cPLS scores\n") 
  scores <- run_cpls(counts, v = n_pls_comp)$scores
  
  if(verbose) cat("\tComputing fdr likelihoods\n")
  likelihoods <- fdr(scores)$likelihood
  sorted_likelihoods <- sort(likelihoods[lower.tri(likelihoods)])
  
  if(is.null(threshold)) {
    threshold <- sorted_likelihoods[min(c(length(sorted_likelihoods), 
                                          max_n_edges + 1))]
  } 
  n_edges <- sum(sorted_likelihoods < threshold)
  if(verbose) cat("\t", n_edges, "edges found using threshold", threshold, "\n")
  if(n_edges > max_n_edges) {
    threshold_new <- sorted_likelihoods[min(c(length(sorted_likelihoods), 
                                              max_n_edges))]
    warning(paste("Default threshold (", threshold, ") results in", n_edges, 
                  "edges. Resetting threshold to", threshold_new, "."))
    threshold <- threshold_new
  }
  
  if(make_plots) {
    if(verbose) cat("\tCreating cytoscape file\n")
    create_cytoscape_file(1 * (likelihoods <= threshold),
                          scores, 
                          colnames(counts), 
                          network_title)
    
    if(verbose) cat("\tCreating histograms for scores and likelihoods\n")
    par(mfrow = c(1, 2))
    hist(scores[lower.tri(scores)], 
         main = network_title, xlab = "scores")
    hist(sorted_likelihoods, 
         main = network_title, xlab = "empirical Bayes f0/f")
  }
  
  if(verbose) cat("\t- End of analysis for network", network_name, "\n")
  
  return(list(network_title = network_title,
              scores = scores,
              likelihoods = likelihoods,
              threshold = threshold,
              genes = colnames(counts)))
}


analyze_differential_network <- function(counts, n_perm = 1000, genes = NULL, 
                                         score_fn = NULL, perm = TRUE, 
                                         boot = NULL, parallel = FALSE) {
  if(is.null(genes)) {
    genes <- colnames(counts[[1]])
  }
  
  if(is.null(score_fn)) {
    score_fn <- function(x) { 
      standardize_scores(run_cpls(x, v = 1)$scores)
      #1 * (fdr(run_cpls(x, v = 1)$scores)$likelihood < significance)
    }
  }
  
  #Permutations for mouse data (sample of size 6)
  if(perm) {
    n1 <- nrow(counts[[1]])
    n2 <- nrow(counts[[2]])
    cat("Enumerating permutations for samples with size", n1, "and", n2, "\n")
    total <- factorial(n1 + n2) / (factorial(n1) * factorial(n2))
    total <- ifelse(n1 == n2, total / 2, total)
    if(total <= 35) {
      #Index all 6!/(2*3!*3!) = 10 combinations.
      perm <- combn(1:(n1 + n2), n1)
      if(n1 == n2) {
        perm <- perm[, 1:total]
      }
      print("All permutations: ")
      print(perm)
    } else {
      cat("Sample size too large to enumerate all permutations")
      perm <- NULL
    }
  }

  results <- test.class.genes(counts[[1]], 
                              counts[[2]], 
                              genelist = genes,
                              scores = score_fn, 
                              num.permutations = n_perm, 
                              perms = perm,
                              boot = boot,
                              parallel = parallel)
  return(results)
}