#' @title Count data partial least squares
#' @description Reconstructs a full genomic association network from a set of 
#'   next-generation sequencing count data.
#' @param x the n by p matrix of raw count data. Rows should correspond to
#'   samples, and columns should correspond to genes.
#' @param v the number of PLS components to compute.
#' @param signif (optional) the signifiance threshold for inferring the 
#'   network through empirical bayes fdr. If NULL, only scores are returned.
#' @return a matrix of association scores. 
run_cpls <- function(x, v = 3, signif = NULL) {
  n <- nrow(x) #Number of observations.
  p <- ncol(x) #Number of genes.
  if(v >= n) {
    stop("v must be less than n (the number of rows in X)")
  }
  if(n > p) {
    warning("expected p > n; matrix should have observations in rows and 
            genes in columns.")
  }
  if(is.data.frame(x)) {
    x <- as.matrix(x)
  }
  if(any(!is.numeric(x))) {
    stop("x contains non-numeric value(s).")
  }
  
  #Total read counts in each sample.
  num_reads_by_sample <- apply(x, 1, sum) 
  
  #The association scores.
  s <- matrix(0, nrow = p, ncol = p)
  
  #Standardization of explanatory variables.
  standardize <- function(x) {
    apply(x, 2, function(x) { (x - mean(x))/sd(x) })
  }
  #If any gene has zero variance, add one count to its first sample.
  zero_variance_genes <- which(apply(x, 2, function(x) var(x) == 0))
  if(length(zero_variance_genes) > 0) {
    cat("Adding one count to first sample for genes:", zero_variance_genes, ".")
    x[1, zero_variance_genes] <- x[1, zero_variance_genes] + 1
  }
  x_std <- standardize(x)
  
  
  #S1-S5.
  for(i in 1:p) {
    if((i %% 1000) == 0) cat("gene", i, "of", p, "\n")
    
    #The PLS coefficients and components.
    pls_coef <- array(0, dim = c(p, v)) 
    pls_comp <- array(0, dim = c(n, v))
    
    #Initialize the deflated x matrix used in each iteration.
    #These explanatory variables ARE standardized.
    x_deflated <- x_std[, -i]
    
    #S2: compute the PLS coefficients and components.
    for(j in 1:v) {
      #The response variable IS standardized.
      pls_coef[-i, j] <- crossprod(x_deflated, x[, i]) # t(x_deflated) %*% x[, i]
      pls_coef[, j]  <- pls_coef[, j]  / sqrt(sum(pls_coef[, j]^2))
      pls_comp[, j] <- x_deflated %*% pls_coef[-i, j]
      
      temp <- tryCatch(solve(crossprod(pls_comp[, j])), 
                       error = function(e) { warning(e); return(NA) })
      #Check that solve() was successful.
      if(!is.na(temp)) {
        x_deflated <- x_deflated - 
          pls_comp[, j] %*% temp %*% crossprod(pls_comp[, j], x_deflated)
      } else {
        #Only stop if fewer than v components have been computed.
        if(j < v) {
          stop(paste("cPLS algorithm halted. Only v =", j, "of", 
                     v, "components computed."))
        }
      }
    }
    
    #S3: fit a negative binomial model: log(E(X)) = pls_comp'beta + log(N).
    #The negative binomial model coefficients.
    beta <- array(0, dim = c(v))
    
    #The response variable is NOT standardized.
    temp_data <- data.frame(y = x[, i], 
                            C = pls_comp, 
                            total = num_reads_by_sample)
    form <- as.formula(paste("y ~", paste(colnames(temp_data[2:(v + 1)]), 
                                          collapse = "+")))
    
    beta <- tryCatch(
      coef(gam(form, offset = log(total), family = nb(), data = temp_data))[-1],
      error = function(e) {
        print(paste(e, "- gene", i, "of", p, "- setting coefs to zero."))
        return(rep(0, v))
      })
    
    #S4: compute the association scores for each gene pair i and k,
    #    k from 1 to p.
    s[, i] <- pls_coef %*% beta
  }
  
  #S6: Set the diagonal of s to 1 and symmetrize s.
  s <- (s + t(s)) * 0.5
  
  #S7: Transform scores to [-1, 1].
  s <- transform_scores(s)
  
  results <- list(scores = s)
  #Add adjacency matrix to output if a significance threshold was given.
  if(!is.null(signif)) {
    cat("Computing adjacency matrix using theshold", signif, "\n")
    results <- c(results, list(adj_matrix = get_adjacency_matrix(s, signif)))
  }
  return(results)
}

#' Transform scores to range from -1 to 1.
#'
#' @param scores a p by p matrix of scores
#' @return a p by p matrix of transformed scores
transform_scores <- function(scores) {
  #S7: transform s so that the scores are in [-1, 1].
  ranges <- range(scores[lower.tri(scores)]) #Ignore diagonal values.
  scores <- 2 * (scores - ranges[1]) / (ranges[2] - ranges[1]) - 1
  
  return(scores)
}

#' @title Determine which associations are significant
#' @description Determines which associations are significant using emperical 
#'   bayes and false discovery rate adjustment.
#' @param scores a symmetric p by p matrix of association scores.
#' @param significance 
#' @param tol
#' @details Under construction.
#' @return A p by p adjacency matrix.
#' @example None.
#' @export
get_adjacency_matrix <- function(scores, significance = 0.1, tol = 1e-16) {
  if(!is.matrix(scores)) {
    stop("scores must be a matrix.")
  } else if(!isSymmetric(scores, tol)) {
    if(any(scores[upper.tri(scores)] != 0)) {
      warning("scores are neither symmetric nor lower triangular.")
    }
  }
  
  p <- nrow(scores) #Number of genes.
  vals <- scores[lower.tri(scores)] #Vector of scores between each gene.
  
  mu.f0 <- mean(vals)
  sigma.f0 <- sd(vals)
  
  #Emperical Bayes FDR.
  f0 <- function(s) {
    dnorm(s, mu.f0, sigma.f0)
  }
  f <- approxfun(density(vals))
  likelihood <- f0(vals)/f(vals)
  
  #Find significant edges.
  adj_matrix <- matrix(0, nrow = p, ncol = p)
  adj_matrix[lower.tri(adj_matrix)] <- 1 * (likelihood < significance)
  adj_matrix <- adj_matrix + t(adj_matrix)
  
  return(adj_matrix)
}

#' @title Determine which associations are significant
#' @description Determines which associations are significant using emperical 
#'   bayes and false discovery rate adjustment.
#' @param scores a symmetric p by p matrix of association scores.
#' @details Under construction.
#' @return A p by p matrix of likelihood.
#' @example None.
#' @export
fdr <- function(scores) {
  if(length(dim(scores)) != 2) {
    stop("scores should be a 2 dimensional matrix or data frame.")
  }
  
  p <- nrow(scores) #Number of genes.
  vals <- scores[lower.tri(scores)] #Vector of scores between each gene.
  
  mu.f0 <- mean(vals)
  sigma.f0 <- sd(vals)
  
  #Emperical Bayes FDR. Save likelihood of each observation.
  likelihood <- matrix(0, nrow = p, ncol = p)
  likelihood[lower.tri(likelihood)] <- 
    dnorm(vals, mu.f0, sigma.f0) / approx(density(vals), xout = vals)$y
  likelihood <- likelihood + t(likelihood)
  diag(likelihood) <- Inf
  
  return(list(likelihood = likelihood))
}
